import pytest

from mimeo import tools
from mimeo.resources.exc import ResourceNotFound


def test_get_resource_existing():
    with tools.get_resource("logging.yaml") as resource:
        assert resource.name.endswith("resources/logging.yaml")


def test_get_resource_non_existing():
    with pytest.raises(ResourceNotFound) as err:
        tools.get_resource("non-existing-file.yaml")

    assert err.value.args[0] == "No such resource: [non-existing-file.yaml]"



import logging
from pathlib import Path

from mimeo.config.mimeo_config import MimeoOutputDetails
from mimeo.consumers import Consumer

logger = logging.getLogger(__name__)


class FileConsumer(Consumer):

    def __init__(self, output_details: MimeoOutputDetails):
        self.directory = output_details.directory_path
        self.output_path_tmplt = f"{self.directory}/{output_details.file_name_tmplt}"
        self.__count = 0

    def consume(self, data: str) -> None:
        logger.fine(f"Consuming data [{data}]")
        if not Path(self.directory).exists():
            logger.info(f"Creating output directory [{self.directory}]")
            Path(self.directory).mkdir(parents=True, exist_ok=True)

        self.__count += 1
        file_name = self.output_path_tmplt.format(self.__count)

        logger.info(f"Writing data into file [{file_name}]")
        with open(file_name, "w") as output_file:
            output_file.write(data)




import shutil
from os import path

import pytest

from mimeo import Mimeograph
from mimeo.config import MimeoConfig
from mimeo.context import MimeoContextManager


@pytest.fixture(autouse=True)
def teardown():
    yield
    # Teardown
    shutil.rmtree("test_mimeograph-dir")


def test_produce():
    config = {
        "output_format": "xml",
        "indent": 4,
        "xml_declaration": True,
        "output_details": {
            "direction": "file",
            "directory_path": "test_mimeograph-dir",
            "file_name": "output"
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }
    mimeo_config = MimeoConfig(config)
    with MimeoContextManager(mimeo_config):
        mimeo = Mimeograph(mimeo_config)

        assert not path.exists("test_mimeograph-dir")
        mimeo.produce()
        assert path.exists("test_mimeograph-dir")
        for i in range(1, 11):
            file_path = f"test_mimeograph-dir/output-{i}.xml"
            assert path.exists(file_path)

            with open(file_path, "r") as file_content:
                assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
                assert file_content.readline() == '<SomeEntity>\n'
                assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
                assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
                assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
                assert file_content.readline() == '</SomeEntity>\n'







from .mimeo_utils import (AutoIncrementUtil, CityUtil, CountryUtil,
                          CurrentIterationUtil, DateTimeUtil, DateUtil,
                          FirstNameUtil, KeyUtil, LastNameUtil, MimeoUtil,
                          RandomIntegerUtil, RandomItemUtil, RandomStringUtil)
from .renderer import MimeoRenderer



from typing.io import TextIO

from mimeo.resources.exc import ResourceNotFound

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources

from mimeo import resources as data


def get_resource(resource_name: str) -> TextIO:
    try:
        return pkg_resources.open_text(data, resource_name)
    except FileNotFoundError:
        raise ResourceNotFound(f"No such resource: [{resource_name}]")



from mimeo.meta.exc import InstanceNotAlive


class Alive:

    def __init__(self):
        self._alive = False

    def __enter__(self):
        self._alive = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._alive = False
        return None

    def assert_alive(self):
        if not self.is_alive():
            raise InstanceNotAlive("The instance is not alive!")
        return self.is_alive()

    def is_alive(self):
        return self._alive


class OnlyOneAlive(type):

    __INSTANCES = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__INSTANCES:
            cls.__INSTANCES[cls] = []

        alive_instance = next((instance for instance in cls.__INSTANCES[cls] if instance.is_alive()), None)
        if alive_instance is not None:
            return alive_instance
        else:
            instance = super().__call__(*args, **kwargs)
            cls.__INSTANCES[cls].append(instance)
            return instance



import pandas

from mimeo import tools
from mimeo.database.exc import InvalidIndex, InvalidSex


class FirstName:

    def __init__(self, name: str, sex: str):
        self.name = name
        self.sex = sex

    def __str__(self) -> str:
        return str({
            "name": self.name,
            "sex": self.sex
        })

    def __repr__(self) -> str:
        return f"FirstName('{self.name}', '{self.sex}')"


class FirstNamesDB:

    NUM_OF_RECORDS = 7455
    __SUPPORTED_SEX = ("M", "F")
    __FIRST_NAMES_DB = "forenames.csv"
    __FIRST_NAMES_DF = None
    __FIRST_NAMES = None
    __NAMES_FOR_SEX = {}

    def get_first_name_at(self, index: int) -> FirstName:
        first_names = self.__get_first_names()
        try:
            return first_names[index]
        except IndexError:
            raise InvalidIndex(f"Provided index [{index}] is out or the range: 0-{FirstNamesDB.NUM_OF_RECORDS-1}!")

    def get_first_names_by_sex(self, sex: str) -> list:
        if sex in FirstNamesDB.__SUPPORTED_SEX:
            return self.__get_first_names_by_sex(sex).copy()
        else:
            raise InvalidSex("Invalid sex (use M or F)!")

    @classmethod
    def get_first_names(cls) -> list:
        return cls.__get_first_names().copy()

    @classmethod
    def __get_first_names_by_sex(cls, sex: str) -> list:
        if sex not in cls.__NAMES_FOR_SEX:
            first_names = cls.__get_first_names()
            cls.__NAMES_FOR_SEX[sex] = list(filter(lambda first_name: first_name.sex == sex, first_names))
        return cls.__NAMES_FOR_SEX[sex]

    @classmethod
    def __get_first_names(cls) -> list:
        if cls.__FIRST_NAMES is None:
            cls.__FIRST_NAMES = [FirstName(row.NAME, row.SEX)
                                 for row in cls.__get_first_names_df().itertuples()]
        return cls.__FIRST_NAMES

    @classmethod
    def __get_first_names_df(cls) -> pandas.DataFrame:
        if cls.__FIRST_NAMES_DF is None:
            cls.__FIRST_NAMES_DF = pandas.read_csv(tools.get_resource(FirstNamesDB.__FIRST_NAMES_DB))
        return cls.__FIRST_NAMES_DF



import random

from mimeo.context import MimeoIteration
from mimeo.context.exc import (ContextIterationNotFound,
                               MinimumIdentifierReached,
                               UninitializedContextIteration)
from mimeo.database import MimeoDB
from mimeo.database.exc import CountryNotFound, OutOfStock


class MimeoContext:

    __ALL = "_ALL_"
    __INITIAL_COUNT = "init-count"
    __INDEXES = "indexes"

    def __init__(self, name: str):
        self.name = name
        self.__id = 0
        self.__iterations = []
        self.__countries_indexes = None
        self.__cities_indexes = {}
        self.__first_names_indexes = {}
        self.__last_names_indexes = None

    def next_id(self) -> int:
        self.__id += 1
        return self.__id

    def curr_id(self) -> int:
        return self.__id

    def prev_id(self) -> int:
        if self.__id > 0:
            self.__id -= 1
            return self.__id
        else:
            raise MinimumIdentifierReached("There's no previous ID!")

    def next_iteration(self) -> MimeoIteration:
        next_iteration_id = 1 if len(self.__iterations) == 0 else self.__iterations[-1].id + 1
        next_iteration = MimeoIteration(next_iteration_id)
        self.__iterations.append(next_iteration)
        return next_iteration

    def curr_iteration(self) -> MimeoIteration:
        if len(self.__iterations) > 0:
            return self.__iterations[-1]
        else:
            raise UninitializedContextIteration(f"No iteration has been initialized for the current context [{self.name}]")

    def get_iteration(self, iteration_id: int) -> MimeoIteration:
        iteration = next(filter(lambda i: i.id == iteration_id, self.__iterations), None)
        if iteration is not None:
            return iteration
        else:
            raise ContextIterationNotFound(f"No iteration with id [{iteration_id}] "
                                           f"has been initialized for the current context [{self.name}]")

    def clear_iterations(self) -> None:
        self.__iterations = []

    def next_country_index(self):
        self.__initialize_countries_indexes()
        self.__validate_countries()

        return self.__countries_indexes.pop()

    def next_city_index(self, country: str = None):
        country = country if country is not None else MimeoContext.__ALL
        self.__initialize_cities_indexes(country)
        self.__validate_cities(country)

        return self.__cities_indexes[country][MimeoContext.__INDEXES].pop()

    def next_first_name_index(self, sex: str = None):
        sex = sex if sex is not None else MimeoContext.__ALL
        self.__initialize_first_names_indexes(sex)
        self.__validate_first_names(sex)

        return self.__first_names_indexes[sex][MimeoContext.__INDEXES].pop()

    def next_last_name_index(self):
        self.__initialize_last_names_indexes()
        self.__validate_last_names()

        return self.__last_names_indexes.pop()

    def __initialize_countries_indexes(self):
        if self.__countries_indexes is None:
            countries_indexes = random.sample(range(MimeoDB.NUM_OF_COUNTRIES), MimeoDB.NUM_OF_COUNTRIES)
            self.__countries_indexes = countries_indexes

    def __initialize_cities_indexes(self, country: str):
        if country not in self.__cities_indexes:
            if country == MimeoContext.__ALL:
                num_of_entries = MimeoDB.NUM_OF_CITIES
            else:
                country_cities = MimeoDB().get_cities_of(country)
                num_of_entries = len(country_cities)
                if num_of_entries == 0:
                    raise CountryNotFound(f"Mimeo database does not contain any cities of provided country [{country}].")

            cities_indexes = random.sample(range(num_of_entries), num_of_entries)
            self.__cities_indexes[country] = {
                MimeoContext.__INITIAL_COUNT: num_of_entries,
                MimeoContext.__INDEXES: cities_indexes
            }

    def __initialize_first_names_indexes(self, sex: str):
        if sex not in self.__first_names_indexes:
            if sex == MimeoContext.__ALL:
                num_of_entries = MimeoDB.NUM_OF_FIRST_NAMES
            else:
                first_names_for_sex = MimeoDB().get_first_names_by_sex(sex)
                num_of_entries = len(first_names_for_sex)

            first_names_indexes = random.sample(range(num_of_entries), num_of_entries)
            self.__first_names_indexes[sex] = {
                MimeoContext.__INITIAL_COUNT: num_of_entries,
                MimeoContext.__INDEXES: first_names_indexes
            }

    def __initialize_last_names_indexes(self):
        if self.__last_names_indexes is None:
            last_names_indexes = random.sample(range(MimeoDB.NUM_OF_LAST_NAMES), MimeoDB.NUM_OF_LAST_NAMES)
            self.__last_names_indexes = last_names_indexes

    def __validate_countries(self) -> None:
        if len(self.__countries_indexes) == 0:
            raise OutOfStock(f"No more unique values, database contain only {MimeoDB.NUM_OF_COUNTRIES} countries.")

    def __validate_cities(self, country: str) -> None:
        if len(self.__cities_indexes[country][MimeoContext.__INDEXES]) == 0:
            init_count = self.__cities_indexes[country][MimeoContext.__INITIAL_COUNT]
            if country == MimeoContext.__ALL:
                raise OutOfStock(f"No more unique values, database contain only {init_count} cities.")
            else:
                raise OutOfStock(f"No more unique values, database contain only {init_count} cities of {country}.")

    def __validate_first_names(self, sex: str) -> None:
        if len(self.__first_names_indexes[sex][MimeoContext.__INDEXES]) == 0:
            init_count = self.__first_names_indexes[sex][MimeoContext.__INITIAL_COUNT]
            if sex == MimeoContext.__ALL:
                raise OutOfStock(f"No more unique values, database contain only {init_count} first names.")
            else:
                raise OutOfStock(f"No more unique values, database contain only {init_count} "
                                 f"{'male' if sex == 'M' else 'female'} first names.")

    def __validate_last_names(self) -> None:
        if len(self.__last_names_indexes) == 0:
            raise OutOfStock(f"No more unique values, database contain only {MimeoDB.NUM_OF_LAST_NAMES} last names.")



import logging.config

import yaml
from haggis.logs import add_logging_level

from mimeo import tools

from .filters import DetailedFilter, RegularFilter


def setup_logging():
    add_logging_level("FINE", logging.DEBUG - 1)
    with tools.get_resource("logging.yaml") as config_file:
        config = yaml.safe_load(config_file.read())
        logging.config.dictConfig(config)



import random
import string
from abc import ABCMeta, abstractmethod
from datetime import date, datetime, timedelta

from mimeo.context import MimeoContext, MimeoContextManager
from mimeo.context.annotations import mimeo_context
from mimeo.database import Country, MimeoDB
from mimeo.database.exc import CountryNotFound, InvalidSex
from mimeo.utils.exc import InvalidValue


class MimeoUtil(metaclass=ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'KEY') and
                hasattr(subclass, 'render') and
                callable(subclass.render) and
                NotImplemented)

    @abstractmethod
    def render(self):
        raise NotImplementedError


class RandomStringUtil(MimeoUtil):

    KEY = "random_str"

    def __init__(self, **kwargs):
        self.__length = kwargs.get("length", 20)

    def render(self):
        return "".join(random.choice(string.ascii_letters) for _ in range(self.__length))


class RandomIntegerUtil(MimeoUtil):

    KEY = "random_int"

    def __init__(self, **kwargs):
        self.__start = kwargs.get("start", 1)
        self.__limit = kwargs.get("limit", 100)

    def render(self):
        return random.randrange(self.__start, self.__limit + 1)


class RandomItemUtil(MimeoUtil):

    KEY = "random_item"

    def __init__(self, **kwargs):
        self.__items = kwargs.get("items", [])

    def render(self):
        length = len(self.__items)
        return "" if length == 0 else self.__items[random.randrange(0, length)]


class DateUtil(MimeoUtil):

    KEY = "date"

    def __init__(self, **kwargs):
        self.__days_delta = kwargs.get("days_delta", 0)

    def render(self):
        date_value = date.today() if self.__days_delta == 0 else date.today() + timedelta(days=self.__days_delta)
        return date_value.strftime("%Y-%m-%d")


class DateTimeUtil(MimeoUtil):

    KEY = "date_time"

    def __init__(self, **kwargs):
        self.__days_delta = kwargs.get("days_delta", 0)
        self.__hours_delta = kwargs.get("hours_delta", 0)
        self.__minutes_delta = kwargs.get("minutes_delta", 0)
        self.__seconds_delta = kwargs.get("seconds_delta", 0)

    def render(self):
        time_value = datetime.now() + timedelta(days=self.__days_delta,
                                                hours=self.__hours_delta,
                                                minutes=self.__minutes_delta,
                                                seconds=self.__seconds_delta)
        return time_value.strftime("%Y-%m-%dT%H:%M:%S")


class AutoIncrementUtil(MimeoUtil):

    KEY = "auto_increment"

    def __init__(self, **kwargs):
        self.__pattern = kwargs.get("pattern", "{:05d}")

    @mimeo_context
    def render(self, context: MimeoContext = None):
        try:
            identifier = context.next_id()
            return self.__pattern.format(identifier)
        except AttributeError:
            context.prev_id()
            raise InvalidValue(f"The {self.KEY} Mimeo Util require a string value for the pattern parameter "
                               f"and was: [{self.__pattern}].")


class CurrentIterationUtil(MimeoUtil):

    KEY = "curr_iter"

    def __init__(self, **kwargs):
        self.__context_name = kwargs.get("context")

    @mimeo_context
    def render(self, context: MimeoContext = None):
        context = context if self.__context_name is None else MimeoContextManager().get_context(self.__context_name)
        return context.curr_iteration().id


class KeyUtil(MimeoUtil):

    KEY = "key"

    def __init__(self, **kwargs):
        self.__context_name = kwargs.get("context")
        self.__iteration = kwargs.get("iteration")

    @mimeo_context
    def render(self, context: MimeoContext = None):
        context = context if self.__context_name is None else MimeoContextManager().get_context(self.__context_name)
        iteration = context.curr_iteration() if self.__iteration is None else context.get_iteration(self.__iteration)
        return iteration.key


class CityUtil(MimeoUtil):

    KEY = "city"
    __MIMEO_DB = MimeoDB()

    def __init__(self, **kwargs):
        self.__unique = kwargs.get("unique", True)
        self.__country = kwargs.get("country", None)

    @mimeo_context
    def render(self, context: MimeoContext = None):
        if self.__country is None:
            if self.__unique:
                index = context.next_city_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_CITIES)
            city = self.__MIMEO_DB.get_city_at(index)
        else:
            country_cities = self.__MIMEO_DB.get_cities_of(self.__country)
            country_cities_count = len(country_cities)
            if country_cities_count == 0:
                raise CountryNotFound(f"Mimeo database does not contain any cities of provided country [{self.__country}].")

            if self.__unique:
                index = context.next_city_index(self.__country)
            else:
                index = random.randrange(country_cities_count)
            city = country_cities[index]

        return city.name_ascii


class CountryUtil(MimeoUtil):

    KEY = "country"

    __VALUE_NAME = "name"
    __VALUE_ISO3 = "iso3"
    __VALUE_ISO2 = "iso2"
    __MIMEO_DB = MimeoDB()

    def __init__(self, **kwargs):
        self.__value = kwargs.get("value", self.__VALUE_NAME)
        self.__unique = kwargs.get("unique", True)
        self.__country = kwargs.get("country", None)

    @mimeo_context
    def render(self, context: MimeoContext = None):
        if self.__value == self.__VALUE_NAME:
            return self.__get_country(context).name
        elif self.__value == self.__VALUE_ISO3:
            return self.__get_country(context).iso_3
        elif self.__value == self.__VALUE_ISO2:
            return self.__get_country(context).iso_2
        else:
            raise InvalidValue(f"The `country` Mimeo Util does not support such value [{self.__value}]. "
                               f"Supported values are: "
                               f"{self.__VALUE_NAME} (default), {self.__VALUE_ISO3}, {self.__VALUE_ISO2}.")

    def __get_country(self, context: MimeoContext) -> Country:
        if self.__country is not None:
            countries = self.__MIMEO_DB.get_countries()
            country_found = next(filter(lambda c: self.__country in [c.name, c.iso_3, c.iso_2], countries), None)
            if country_found is not None:
                return country_found
            else:
                raise CountryNotFound(f"Mimeo database does not contain such a country [{self.__country}].")
        else:
            if self.__unique:
                index = context.next_country_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_COUNTRIES)

            country = self.__MIMEO_DB.get_country_at(index)
            return country


class FirstNameUtil(MimeoUtil):

    KEY = "first_name"
    __MIMEO_DB = MimeoDB()

    def __init__(self, **kwargs):
        self.__unique = kwargs.get("unique", True)
        self.__sex = self._standardize_sex(kwargs.get("sex"))

    @mimeo_context
    def render(self, context: MimeoContext = None):
        if self.__sex is None:
            if self.__unique:
                index = context.next_first_name_index()
            else:
                index = random.randrange(MimeoDB.NUM_OF_FIRST_NAMES)
            first_name = self.__MIMEO_DB.get_first_name_at(index)
        else:
            first_name_for_sex = self.__MIMEO_DB.get_first_names_by_sex(self.__sex)
            first_name_for_sex_count = len(first_name_for_sex)

            if self.__unique:
                index = context.next_first_name_index(self.__sex)
            else:
                index = random.randrange(first_name_for_sex_count)
            first_name = first_name_for_sex[index]

        return first_name.name

    @classmethod
    def _standardize_sex(cls, sex: str):
        if sex is None:
            return sex
        elif sex.upper() in ["M", "MALE"]:
            return "M"
        elif sex.upper() in ["F", "FEMALE"]:
            return "F"
        else:
            raise InvalidSex("Invalid sex (use M, F, Male or Female)!")


class LastNameUtil(MimeoUtil):

    KEY = "last_name"
    __MIMEO_DB = MimeoDB()

    def __init__(self, **kwargs):
        self.__unique = kwargs.get("unique", True)

    @mimeo_context
    def render(self, context: MimeoContext = None):
        if self.__unique:
            index = context.next_first_name_index()
        else:
            index = random.randrange(MimeoDB.NUM_OF_FIRST_NAMES)
        last_name = self.__MIMEO_DB.get_last_name_at(index)

        return last_name



import functools

from mimeo.config.mimeo_config import MimeoTemplate
from mimeo.context import MimeoContext, MimeoContextManager


def mimeo_context(func):
    @functools.wraps(func)
    def inject_context(*args, **kwargs):
        if any(isinstance(arg, MimeoContext) for arg in args) or "context" in kwargs:
            result = func(*args, **kwargs)
        else:
            result = func(*args, **kwargs, context=MimeoContextManager().get_current_context())
        return result

    return inject_context


def mimeo_context_switch(func):
    @functools.wraps(func)
    def switch_context(*args, **kwargs):
        context_mng = MimeoContextManager()
        prev_context = context_mng.get_current_context()

        if "template" in kwargs:
            template = kwargs["template"]
        else:
            template = next(arg for arg in args if isinstance(arg, MimeoTemplate))
        context_name = template.model.context_name
        next_context = context_mng.get_context(context_name)
        context_mng.set_current_context(next_context)
        result = func(*args, **kwargs)

        context_mng.set_current_context(prev_context)
        return result

    return switch_context


def mimeo_next_iteration(func):
    @functools.wraps(func)
    def next_iteration(*args, **kwargs):
        MimeoContextManager().get_current_context().next_iteration()
        result = func(*args, **kwargs)
        return result

    return next_iteration


def mimeo_clear_iterations(func):
    @functools.wraps(func)
    def clear_iterations(*args, **kwargs):
        MimeoContextManager().get_current_context().clear_iterations()
        result = func(*args, **kwargs)
        return result

    return clear_iterations



from mimeo.database import (CitiesDB, City, CountriesDB, Country, FirstName,
                            FirstNamesDB, LastNamesDB)


class MimeoDB:

    NUM_OF_CITIES = CitiesDB.NUM_OF_RECORDS
    NUM_OF_COUNTRIES = CountriesDB.NUM_OF_RECORDS
    NUM_OF_FIRST_NAMES = FirstNamesDB.NUM_OF_RECORDS
    NUM_OF_LAST_NAMES = LastNamesDB.NUM_OF_RECORDS

    def __init__(self):
        self.__cities_db = CitiesDB()
        self.__countries_db = CountriesDB()
        self.__first_names_db = FirstNamesDB()
        self.__last_names_db = LastNamesDB()

    def get_cities(self) -> list:
        return self.__cities_db.get_cities()

    def get_city_at(self, index: int) -> City:
        return self.__cities_db.get_city_at(index)

    def get_cities_of(self, country: str) -> list:
        country = next(filter(lambda c: country in [c.iso_3, c.iso_2, c.name], self.get_countries()), None)
        if country is None:
            return []
        else:
            return self.__cities_db.get_cities_of(country.iso_3)

    def get_countries(self) -> list:
        return self.__countries_db.get_countries()

    def get_country_at(self, index: int) -> Country:
        return self.__countries_db.get_country_at(index)

    def get_country_by_iso_3(self, iso_3: str) -> Country:
        return self.__countries_db.get_country_by_iso_3(iso_3)

    def get_country_by_iso_2(self, iso_2: str) -> Country:
        return self.__countries_db.get_country_by_iso_2(iso_2)

    def get_country_by_name(self, name: str) -> Country:
        return self.__countries_db.get_country_by_name(name)

    def get_first_names(self) -> list:
        return self.__first_names_db.get_first_names()

    def get_first_name_at(self, index: int) -> FirstName:
        return self.__first_names_db.get_first_name_at(index)

    def get_first_names_by_sex(self, sex: str) -> list:
        return self.__first_names_db.get_first_names_by_sex(sex)

    def get_last_names(self) -> list:
        return self.__last_names_db.get_last_names()

    def get_last_name_at(self, index: int) -> FirstName:
        return self.__last_names_db.get_last_name_at(index)



import json
import logging
import shutil
import sys
from base64 import b64encode
from glob import glob
from http import HTTPStatus
from os import listdir, path, remove, rmdir
from pathlib import Path

import pytest
import responses
from responses import matchers

import mimeo.__main__ as MimeoCLI
from mimeo.config.exc import MissingRequiredProperty
from mimeo.exc import EnvironmentNotFound, EnvironmentsFileNotFound


@pytest.fixture(autouse=True)
def minimum_config():
    return {
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }


@pytest.fixture(autouse=True)
def default_config():
    return {
        "output_format": "xml",
        "indent": 4,
        "xml_declaration": True,
        "output_details": {
            "direction": "file",
            "directory_path": "test_mimeo_cli-dir/output",
            "file_name": "output-file"
        },
        "_templates_": [
            {
                "count": 10,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }


@pytest.fixture(autouse=True)
def http_config():
    return {
        "output_details": {
            "direction": "http",
            "method": "POST",
            "protocol": "http",
            "host": "localhost",
            "port": 8080,
            "endpoint": "/document",
            "auth": "digest",
            "username": "admin",
            "password": "admin"
        },
        "_templates_": [
            {
                "count": 1,
                "model": {
                    "SomeEntity": {
                        "ChildNode1": 1,
                        "ChildNode2": "value-2",
                        "ChildNode3": True
                    }
                }
            }
        ]
    }


@pytest.fixture(autouse=True)
def http_default_envs():
    return {
        "default": {
            "protocol": "https",
            "host": "11.111.11.111",
            "port": 8000,
            "auth": "basic",
            "username": "custom-username",
            "password": "custom-password"
        }
    }


@pytest.fixture(autouse=True)
def http_custom_envs():
    return {
        "custom": {
            "protocol": "https",
            "host": "11.111.11.111",
            "port": 8000,
            "auth": "basic",
            "username": "custom-username",
            "password": "custom-password"
        }
    }


@pytest.fixture(autouse=True)
def setup_and_teardown(minimum_config, default_config, http_config, http_default_envs, http_custom_envs):
    # Setup
    Path("test_mimeo_cli-dir").mkdir(parents=True, exist_ok=True)
    with open("test_mimeo_cli-dir/minimum-config.json", "w") as file:
        json.dump(minimum_config, file)
    with open("test_mimeo_cli-dir/default-config.json", "w") as file:
        json.dump(default_config, file)
    with open("test_mimeo_cli-dir/http-config.json", "w") as file:
        json.dump(http_config, file)
    with open("mimeo.envs.json", "w") as file:
        json.dump(http_default_envs, file)
    with open("custom-mimeo-envs-file.json", "w") as file:
        json.dump(http_custom_envs, file)

    yield

    # Teardown
    shutil.rmtree("test_mimeo_cli-dir")
    remove("mimeo.envs.json")
    remove("custom-mimeo-envs-file.json")
    for filename in glob("mimeo-output/customized-output-file*"):
        remove(filename)
    for filename in glob("mimeo-output/mimeo-output*"):
        remove(filename)
    if path.exists("mimeo-output") and not listdir("mimeo-output"):
        rmdir("mimeo-output")


def test_basic_use():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


@responses.activate
def test_directory_path():
    sys.argv = ["mimeo", "test_mimeo_cli-dir"]

    responses.add(responses.POST, "http://localhost:8080/document")
    assert not path.exists("test_mimeo_cli-dir/output")
    assert not path.exists("mimeo-output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'

    assert path.exists("mimeo-output")
    for i in range(1, 11):
        file_path = f"mimeo-output/mimeo-output-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<SomeEntity>' \
                                              '<ChildNode1>1</ChildNode1>' \
                                              '<ChildNode2>value-2</ChildNode2>' \
                                              '<ChildNode3>true</ChildNode3>' \
                                              '</SomeEntity>' \


def test_custom_short_xml_declaration_false():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-x", "false"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_xml_declaration_true():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-x", "true"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_xml_declaration_false():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--xml-declaration", "false"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_xml_declaration_true():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--xml-declaration", "true"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_indent_non_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-i", "2"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '  <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '  <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '  <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_short_indent_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-i", "0"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == "<?xml version='1.0' encoding='utf-8'?>\n"
            assert file_content.readline() == '<SomeEntity>' \
                                              '<ChildNode1>1</ChildNode1>' \
                                              '<ChildNode2>value-2</ChildNode2>' \
                                              '<ChildNode3>true</ChildNode3>' \
                                              '</SomeEntity>'


def test_custom_long_indent_non_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--indent", "2"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '  <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '  <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '  <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_indent_zero():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--indent", "0"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == "<?xml version='1.0' encoding='utf-8'?>\n"
            assert file_content.readline() == '<SomeEntity>' \
                                              '<ChildNode1>1</ChildNode1>' \
                                              '<ChildNode2>value-2</ChildNode2>' \
                                              '<ChildNode3>true</ChildNode3>' \
                                              '</SomeEntity>'


def test_custom_short_output_direction():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-o", "stdout"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")


def test_custom_long_output_direction():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--output", "stdout"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")


def test_custom_output_direction_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "stdout"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


def test_custom_short_output_directory_path():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-d", "test_mimeo_cli-dir/customized-output"]

    assert not path.exists("test_mimeo_cli-dir/output")
    assert not path.exists("test_mimeo_cli-dir/customized-output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")
    assert path.exists("test_mimeo_cli-dir/customized-output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/customized-output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_output_directory_path():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--directory", "test_mimeo_cli-dir/customized-output"]

    assert not path.exists("test_mimeo_cli-dir/output")
    assert not path.exists("test_mimeo_cli-dir/customized-output")

    MimeoCLI.main()

    assert not path.exists("test_mimeo_cli-dir/output")
    assert path.exists("test_mimeo_cli-dir/customized-output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/customized-output/output-file-{i}.xml"
        assert path.exists(file_path)

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_output_directory_path_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-d", "test_mimeo_cli-dir/customized-output"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


def test_custom_short_output_file_name():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "-f", "customized-output-file"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/customized-output-file-{i}.xml"
        assert path.exists(file_path)
        assert not path.exists(f"test_mimeo_cli-dir/output/output-file-{i}.xml")

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_long_output_file_name():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--file", "customized-output-file"]

    assert not path.exists("test_mimeo_cli-dir/output")

    MimeoCLI.main()

    assert path.exists("test_mimeo_cli-dir/output")
    for i in range(1, 11):
        file_path = f"test_mimeo_cli-dir/output/customized-output-file-{i}.xml"
        assert path.exists(file_path)
        assert not path.exists(f"test_mimeo_cli-dir/output/output-file-{i}.xml")

        with open(file_path, "r") as file_content:
            assert file_content.readline() == '<?xml version="1.0" encoding="utf-8"?>\n'
            assert file_content.readline() == '<SomeEntity>\n'
            assert file_content.readline() == '    <ChildNode1>1</ChildNode1>\n'
            assert file_content.readline() == '    <ChildNode2>value-2</ChildNode2>\n'
            assert file_content.readline() == '    <ChildNode3>true</ChildNode3>\n'
            assert file_content.readline() == '</SomeEntity>\n'


def test_custom_output_file_name_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-f", "customized-output-file"]

    try:
        MimeoCLI.main()
    except KeyError:
        assert False


@responses.activate
def test_custom_short_output_http_host():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "-H", "198.168.1.1"]

    responses.add(responses.POST, "http://198.168.1.1:8080/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_output_http_host():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-host", "198.168.1.1"]

    responses.add(responses.POST, "http://198.168.1.1:8080/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_host_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-H", "198.168.1.1"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_short_output_http_port():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "-p", "8081"]

    responses.add(responses.POST, "http://localhost:8081/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_output_http_port():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-port", "8081"]

    responses.add(responses.POST, "http://localhost:8081/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_port_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-p", "8081"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_short_output_http_endpoint():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "-E", "/v2/document"]

    responses.add(responses.POST, "http://localhost:8080/v2/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_output_http_endpoint():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-endpoint", "/v2/document"]

    responses.add(responses.POST, "http://localhost:8080/v2/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_endpoint_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-E", "/v2/document"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_short_output_http_username():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-auth", "basic", "-U", "custom-user"]

    responses.add(
        responses.POST,
        "http://localhost:8080/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("custom-user", "admin")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_output_http_username():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-auth", "basic", "--http-user", "custom-user"]

    responses.add(
        responses.POST,
        "http://localhost:8080/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("custom-user", "admin")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_username_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-U", "custom-user"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_short_output_http_password():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-auth", "basic", "-P", "custom-password"]

    responses.add(
        responses.POST,
        "http://localhost:8080/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("admin", "custom-password")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_output_http_password():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json",
                "--http-auth", "basic",
                "--http-password", "custom-password"]

    responses.add(
        responses.POST,
        "http://localhost:8080/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("admin", "custom-password")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_password_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-P", "custom-password"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_long_output_http_method():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-method", "PUT"]

    responses.add(responses.PUT, "http://localhost:8080/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_method_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "--http-method", "PUT"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_long_output_http_protocol():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-protocol", "https"]

    responses.add(responses.POST, "https://localhost:8080/document", json={"success": True}, status=HTTPStatus.OK)
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_protocol_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "--http-protocol", "https"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_long_output_http_auth():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-auth", "basic"]

    responses.add(
        responses.POST,
        "http://localhost:8080/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("admin", "admin")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_output_http_auth_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "--http-auth", "basic"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


@responses.activate
def test_custom_short_env():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "-e", "default"]

    responses.add(
        responses.POST,
        "https://11.111.11.111:8000/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("custom-username", "custom-password")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_long_env():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json", "--http-env", "default"]

    responses.add(
        responses.POST,
        "https://11.111.11.111:8000/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("custom-username", "custom-password")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


@responses.activate
def test_custom_env_file():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json",
                "--http-envs-file", "custom-mimeo-envs-file.json",
                "-e", "custom"]

    responses.add(
        responses.POST,
        "https://11.111.11.111:8000/document",
        json={"success": True},
        status=HTTPStatus.OK,
        match=[matchers.header_matcher({'Authorization': _generate_authorization("custom-username", "custom-password")})]
    )
    MimeoCLI.main()
    # would throw a ConnectionError when any request call doesn't match registered mocks


def test_custom_env_file_does_throw_error_when_file_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/http-config.json",
                "--http-envs-file", "non-existing-environments-file.json",
                "-e", "custom"]

    with pytest.raises(EnvironmentsFileNotFound) as err:
        MimeoCLI.main()

    assert err.value.args[0] == "Environments file not found [non-existing-environments-file.json]"


def test_custom_env_does_not_throw_key_error_when_output_details_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-e", "default"]

    try:
        MimeoCLI.main()
    except MissingRequiredProperty:
        assert True
    except KeyError:
        assert False


def test_custom_env_does_throw_error_when_environment_does_not_exist():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/minimum-config.json", "-o", "http", "-e", "non-existing"]

    with pytest.raises(EnvironmentNotFound) as err:
        MimeoCLI.main()

    assert err.value.args[0] == "No such env [non-existing] in environments file [mimeo.envs.json]"


def test_logging_mode_default():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.INFO


def test_logging_mode_silent():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--silent"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.WARNING


def test_logging_mode_debug():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--debug"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.DEBUG


def test_logging_mode_fine():
    sys.argv = ["mimeo", "test_mimeo_cli-dir/default-config.json", "--fine"]
    logger = logging.getLogger("mimeo")

    MimeoCLI.main()

    assert logger.getEffectiveLevel() == logging.FINE


def _generate_authorization(username: str, password: str):
    token = b64encode(f"{username}:{password}".encode('utf-8')).decode("ascii")
    return f'Basic {token}'



from mimeo.config import MimeoConfig
from mimeo.context import MimeoContext
from mimeo.context.exc import VarNotFound
from mimeo.meta import Alive, OnlyOneAlive


class MimeoContextManager(Alive, metaclass=OnlyOneAlive):

    __INSTANCES = []

    def __init__(self, mimeo_config: MimeoConfig = None):
        super().__init__()
        self.__mimeo_config = mimeo_config
        self.__vars = {}
        self.__contexts = {}
        self.__current_context = None

    def __enter__(self):
        super().__enter__()
        self.__vars = self.__mimeo_config.vars
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self.__vars = None
        self.__contexts = None
        return None

    def get_context(self, context: str) -> MimeoContext:
        super().assert_alive()
        if context not in self.__contexts:
            self.__contexts[context] = MimeoContext(context)
        return self.__contexts[context]

    def get_current_context(self) -> MimeoContext:
        super().assert_alive()
        return self.__current_context

    def set_current_context(self, context: MimeoContext) -> None:
        super().assert_alive()
        self.__current_context = context

    def get_var(self, variable_name: str):
        value = self.__vars.get(variable_name)
        if value is not None:
            return value
        else:
            raise VarNotFound(f"Provided variable [{variable_name}] is not defined!")



from mimeo.config.exc import UnsupportedPropertyValue
from mimeo.config.mimeo_config import MimeoConfig
from mimeo.generators import Generator, XMLGenerator


class GeneratorFactory:

    XML = MimeoConfig.OUTPUT_FORMAT_XML

    @staticmethod
    def get_generator(mimeo_config: MimeoConfig) -> Generator:
        output_format = mimeo_config.output_format
        if output_format == GeneratorFactory.XML:
            return XMLGenerator(mimeo_config)
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_FORMAT_KEY,
                                           output_format,
                                           MimeoConfig.SUPPORTED_OUTPUT_FORMATS)



from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import ConsumerFactory
from mimeo.generators import GeneratorFactory


class Mimeograph:

    def __init__(self, mimeo_config: MimeoConfig):
        self.mimeo_config = mimeo_config
        self.__generator = GeneratorFactory.get_generator(self.mimeo_config)
        self.__consumer = ConsumerFactory.get_consumer(self.mimeo_config)

    def produce(self):
        for data in self.__generator.generate(self.mimeo_config.templates):
            data_str = self.__generator.stringify(data, self.mimeo_config)
            self.__consumer.consume(data_str)



from logging import DEBUG, INFO, Filter, LogRecord


class RegularFilter(Filter):

    def filter(self, record: LogRecord):
        return record.levelno >= INFO


class DetailedFilter(Filter):

    def filter(self, record: LogRecord):
        return record.levelno <= DEBUG



import pandas

from mimeo import tools
from mimeo.database.exc import InvalidIndex


class City:

    def __init__(self, identifier: str, name: str, name_ascii: str, country: str):
        self.id = int(identifier)
        self.name = name
        self.name_ascii = name_ascii
        self.country = country

    def __str__(self) -> str:
        return str({
            "id": self.id,
            "name": self.name,
            "name_ascii": self.name_ascii,
            "country": self.country
        })

    def __repr__(self) -> str:
        return f"City('{self.id}', '{self.name}', '{self.name_ascii}', '{self.country}')"


class CitiesDB:

    NUM_OF_RECORDS = 42905
    __CITIES_DB = "cities.csv"
    __CITIES_DF = None
    __CITIES = None
    __COUNTRY_CITIES = {}

    def get_city_at(self, index: int) -> City:
        cities = self.__get_cities()
        try:
            return cities[index]
        except IndexError:
            raise InvalidIndex(f"Provided index [{index}] is out or the range: 0-{CitiesDB.NUM_OF_RECORDS-1}!")

    def get_cities_of(self, country_iso3: str) -> list:
        return self.__get_country_cities(country_iso3).copy()

    @classmethod
    def get_cities(cls) -> list:
        return cls.__get_cities().copy()

    @classmethod
    def __get_country_cities(cls, country_iso3: str) -> list:
        if country_iso3 not in cls.__COUNTRY_CITIES:
            cities = cls.__get_cities()
            cls.__COUNTRY_CITIES[country_iso3] = list(filter(lambda city: city.country == country_iso3, cities))
        return cls.__COUNTRY_CITIES[country_iso3]

    @classmethod
    def __get_cities(cls) -> list:
        if cls.__CITIES is None:
            cls.__CITIES = [City(row.ID, row.CITY, row.CITY_ASCII, row.COUNTRY)
                            for row in cls.__get_cities_df().itertuples()]
        return cls.__CITIES

    @classmethod
    def __get_cities_df(cls) -> pandas.DataFrame:
        if cls.__CITIES_DF is None:
            cls.__CITIES_DF = pandas.read_csv(tools.get_resource(CitiesDB.__CITIES_DB))
        return cls.__CITIES_DF



from mimeo import tools
from mimeo.database.exc import InvalidIndex


class LastNamesDB:

    NUM_OF_RECORDS = 151670
    __LAST_NAMES_DB = "surnames.txt"
    __LAST_NAMES = None

    def get_last_name_at(self, index: int) -> str:
        last_names = self.__get_last_names()
        try:
            return last_names[index]
        except IndexError:
            raise InvalidIndex(f"Provided index [{index}] is out or the range: 0-{LastNamesDB.NUM_OF_RECORDS-1}!")

    @classmethod
    def get_last_names(cls) -> list:
        return cls.__get_last_names().copy()

    @classmethod
    def __get_last_names(cls) -> list:
        if cls.__LAST_NAMES is None:
            with tools.get_resource(LastNamesDB.__LAST_NAMES_DB) as last_names:
                cls.__LAST_NAMES = [line.rstrip() for line in last_names.readlines()]
        return cls.__LAST_NAMES



import pandas

from mimeo import tools
from mimeo.database.exc import InvalidIndex


class Country:

    def __init__(self, iso_3: str, iso_2: str, name: str):
        self.iso_3 = iso_3
        self.iso_2 = iso_2
        self.name = name

    def __str__(self) -> str:
        return str({
            "iso_3": self.iso_3,
            "iso_2": self.iso_2,
            "name": self.name
        })

    def __repr__(self) -> str:
        return f"Country('{self.iso_3}', '{self.iso_2}', '{self.name}')"


class CountriesDB:

    NUM_OF_RECORDS = 239
    __COUNTRIES_DB = "countries.csv"
    __COUNTRIES_DF = None
    __COUNTRIES = None

    def get_country_at(self, index: int) -> Country:
        countries = self.__get_countries()
        try:
            return countries[index]
        except IndexError:
            raise InvalidIndex(f"Provided index [{index}] is out or the range: 0-{len(countries)-1}!")

    def get_country_by_iso_3(self, iso_3: str) -> Country:
        countries = self.__get_countries()
        return next(filter(lambda country: country.iso_3 == iso_3, countries), None)

    def get_country_by_iso_2(self, iso_2: str) -> Country:
        countries = self.__get_countries()
        return next(filter(lambda country: country.iso_2 == iso_2, countries), None)

    def get_country_by_name(self, name: str) -> Country:
        countries = self.__get_countries()
        return next(filter(lambda country: country.name == name, countries), None)

    @classmethod
    def get_countries(cls) -> list:
        return cls.__get_countries().copy()

    @classmethod
    def __get_countries(cls) -> list:
        if cls.__COUNTRIES is None:
            cls.__COUNTRIES = [Country(row.ISO_3, row.ISO_2, row.NAME)
                               for row in cls.__get_countries_df().itertuples()]
        return cls.__COUNTRIES

    @classmethod
    def __get_countries_df(cls) -> pandas.DataFrame:
        if cls.__COUNTRIES_DF is None:
            cls.__COUNTRIES_DF = pandas.read_csv(tools.get_resource(CountriesDB.__COUNTRIES_DB))
        return cls.__COUNTRIES_DF



import json
import logging
from argparse import ArgumentParser
from os import path, walk

from mimeo import Mimeograph
from mimeo.config import MimeoConfig
from mimeo.context import MimeoContextManager
from mimeo.exc import EnvironmentNotFound, EnvironmentsFileNotFound
from mimeo.logging import setup_logging

setup_logging()
logger = logging.getLogger(__name__)

DEFAULT_ENVS_FILE_PATH = "mimeo.envs.json"


class MimeoArgumentParser(ArgumentParser):

    def __init__(self):
        super().__init__(
            prog="mimeo",
            description="Generate data based on a template")
        self.add_argument(
            "-v",
            "--version",
            action="version",
            version="%(prog)s v0.4.1")
        self.add_argument(
            "paths",
            nargs="+",
            type=str,
            help="take paths to Mimeo Configurations")

        mimeo_config_args = self.add_argument_group("Mimeo Configuration arguments")
        mimeo_config_args.add_argument(
            "-x",
            "--xml-declaration",
            type=str,
            choices=["true", "false"],
            help="overwrite the xml_declaration property")
        mimeo_config_args.add_argument(
            "-i",
            "--indent",
            type=int,
            help="overwrite the indent property")
        mimeo_config_args.add_argument(
            "-o",
            "--output",
            type=str,
            choices=["file", "stdout", "http"],
            help="overwrite the output_details/direction property")
        mimeo_config_args.add_argument(
            "-d",
            "--directory",
            type=str,
            metavar="DIRECTORY_PATH",
            help="overwrite the output_details/directory_path property")
        mimeo_config_args.add_argument(
            "-f",
            "--file",
            type=str,
            metavar="FILE_NAME",
            help="overwrite the output_details/file_name property")
        mimeo_config_args.add_argument(
            "-H",
            "--http-host",
            type=str,
            metavar="HOST",
            help="overwrite the output_details/host property")
        mimeo_config_args.add_argument(
            "-p",
            "--http-port",
            type=str,
            metavar="PORT",
            help="overwrite the output_details/port property")
        mimeo_config_args.add_argument(
            "-E",
            "--http-endpoint",
            type=str,
            metavar="ENDPOINT",
            help="overwrite the output_details/endpoint property")
        mimeo_config_args.add_argument(
            "-U",
            "--http-user",
            type=str,
            metavar="USERNAME",
            help="overwrite the output_details/username property")
        mimeo_config_args.add_argument(
            "-P",
            "--http-password",
            type=str,
            metavar="PASSWORD",
            help="overwrite the output_details/password property")
        mimeo_config_args.add_argument(
            "--http-method",
            type=str,
            metavar="METHOD",
            help="overwrite the output_details/method property")
        mimeo_config_args.add_argument(
            "--http-protocol",
            type=str,
            metavar="PROTOCOL",
            help="overwrite the output_details/protocol property")
        mimeo_config_args.add_argument(
            "--http-auth",
            type=str,
            metavar="AUTH",
            help="overwrite the output_details/auth property")
        mimeo_config_args.add_argument(
            "-e",
            "--http-env",
            type=str,
            metavar="ENVIRONMENT",
            help="overwrite the output_details http properties using a mimeo environment configuration")
        mimeo_config_args.add_argument(
            "--http-envs-file",
            type=str,
            metavar="PATH",
            help=f"use a custom environments file (by default: {DEFAULT_ENVS_FILE_PATH})")

        logging_args = self.add_argument_group("Logging arguments")
        logging_args_excl = logging_args.add_mutually_exclusive_group()
        logging_args_excl.add_argument(
            "--silent",
            action="store_true",
            help="disable INFO logs")
        logging_args_excl.add_argument(
            "--debug",
            action="store_true",
            help="enable DEBUG mode")
        logging_args_excl.add_argument(
            "--fine",
            action="store_true",
            help="enable FINE mode")


def main():
    mimeo_parser = MimeoArgumentParser()
    args = mimeo_parser.parse_args()
    customize_logging_level(args)

    logger.info("Starting Mimeo job")
    paths = get_paths(args.paths)
    for config_path in paths:
        logger.info(f"Data generation from Mimeo Config: {config_path}")
        mimeo_config = get_config(config_path, args)
        with MimeoContextManager(mimeo_config):
            Mimeograph(mimeo_config).produce()


def customize_logging_level(args):
    if args.silent:
        logging.getLogger("mimeo").setLevel(logging.WARNING)
    elif args.debug:
        logging.getLogger("mimeo").setLevel(logging.DEBUG)
    elif args.fine:
        logging.getLogger("mimeo").setLevel(logging.FINE)


def get_paths(paths: list) -> list:
    file_paths = []
    for file_path in paths:
        if path.isdir(file_path):
            for dir_path, _, file_names in walk(file_path):
                for file_name in file_names:
                    paths.append(f"{dir_path}/{file_name}")
        elif path.isfile(file_path):
            file_paths.append(file_path)
    return file_paths


def get_config(config_path, args):
    with open(config_path) as config_file:
        config = json.load(config_file)
        if args.http_env is not None:
            envs_file = args.http_envs_file if args.http_envs_file is not None else DEFAULT_ENVS_FILE_PATH
            customize_output_details_with_env(config, envs_file, args.http_env)
        if args.xml_declaration is not None:
            xml_declaration = args.xml_declaration.lower() == "true"
            logger.fine(f"Overwriting xml_declaration to [{xml_declaration}]")
            config[MimeoConfig.XML_DECLARATION_KEY] = xml_declaration
        if args.indent is not None:
            logger.fine(f"Overwriting indent to [{args.indent}]")
            config[MimeoConfig.INDENT_KEY] = args.indent
        if args.output is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY, args.output)
        if args.directory is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_DIRECTORY_PATH_KEY, args.directory)
        if args.file is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_FILE_NAME_KEY, args.file)
        if args.http_method is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_METHOD_KEY, args.http_method)
        if args.http_protocol is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY, args.http_protocol)
        if args.http_auth is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_AUTH_KEY, args.http_auth)
        if args.http_host is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_HOST_KEY, args.http_host)
        if args.http_port is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_PORT_KEY, args.http_port)
        if args.http_endpoint is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_ENDPOINT_KEY, args.http_endpoint)
        if args.http_user is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY, args.http_user)
        if args.http_password is not None:
            customize_output_details(config, MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY, args.http_password)
    mimeo_config = MimeoConfig(config)
    logger.debug(f"Mimeo Config: {mimeo_config}")
    return mimeo_config


def customize_output_details_with_env(config, envs_path, env_name):
    if path.exists(envs_path):
        with open(envs_path) as envs_file:
            envs = json.load(envs_file)
            if env_name in envs:
                env = envs[env_name]
                logger.debug(f"Using environment [{env_name}] from file [{envs_path}]: [{env}]")
                for prop in [MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY,
                             MimeoConfig.OUTPUT_DETAILS_HOST_KEY,
                             MimeoConfig.OUTPUT_DETAILS_PORT_KEY,
                             MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                             MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY,
                             MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY]:
                    prop_value = env.get(prop)
                    if prop_value is not None:
                        customize_output_details(config, prop, prop_value)
            else:
                raise EnvironmentNotFound(f"No such env [{env_name}] in environments file [{envs_path}]")
    else:
        raise EnvironmentsFileNotFound(f"Environments file not found [{envs_path}]")


def customize_output_details(config, key, value):
    if config.get(MimeoConfig.OUTPUT_DETAILS_KEY) is None:
        config[MimeoConfig.OUTPUT_DETAILS_KEY] = {}

    logger.fine(f"Overwriting output details' {key} to [{value}]")
    config[MimeoConfig.OUTPUT_DETAILS_KEY][key] = value


if __name__ == '__main__':
    main()



class MinimumIdentifierReached(Exception):
    pass


class UninitializedContextIteration(Exception):
    pass


class ContextIterationNotFound(Exception):
    pass


class InvalidSpecialFieldName(Exception):
    pass


class InvalidSpecialFieldValue(Exception):
    pass


class SpecialFieldNotFound(Exception):
    pass


class VarNotFound(Exception):
    pass



import uuid

from mimeo.context.exc import (InvalidSpecialFieldName,
                               InvalidSpecialFieldValue, SpecialFieldNotFound)


class MimeoIteration:

    def __init__(self, identifier: int):
        self.id = identifier
        self.key = str(uuid.uuid4())
        self.__special_fields = {}

    def add_special_field(self, field_name: str, field_value) -> None:
        if not isinstance(field_name, str):
            raise InvalidSpecialFieldName("A special field name needs to be a string value!")
        if isinstance(field_value, dict) or isinstance(field_value, list):
            raise InvalidSpecialFieldValue(f"Provided field value [{field_value}] is invalid (use any atomic value)!")

        self.__special_fields[field_name] = field_value

    def get_special_field(self, field_name: str):
        if field_name not in self.__special_fields:
            raise SpecialFieldNotFound(f"Special Field [{field_name}] has not been found!")

        return self.__special_fields.get(field_name)



from abc import ABCMeta, abstractmethod


class Consumer(metaclass=ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'consume') and
                callable(subclass.consume) and
                NotImplemented)

    @abstractmethod
    def consume(self, data: str):
        raise NotImplementedError



import xml.etree.ElementTree as ElemTree
from abc import ABCMeta, abstractmethod
from typing import Any, Iterator, Union

from mimeo.config.mimeo_config import MimeoTemplate


class Generator(metaclass=ABCMeta):

    __GENERATOR_UTILS_CALL = "GeneratorUtils.get_for_context('{}').{}"

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'generate') and
                callable(subclass.generate) and
                hasattr(subclass, 'stringify') and
                callable(subclass.stringify) or
                NotImplemented)

    @abstractmethod
    def generate(self,
                 templates: Union[list, Iterator[MimeoTemplate]],
                 parent: Any = None) -> Iterator[ElemTree.Element]:
        raise NotImplementedError

    @abstractmethod
    def stringify(self, data, mimeo_config) -> str:
        raise NotImplementedError



import re

from mimeo.config.exc import (IncorrectMimeoConfig, IncorrectMimeoModel,
                              IncorrectMimeoTemplate, InvalidIndent,
                              InvalidVars, MissingRequiredProperty,
                              UnsupportedPropertyValue)
from mimeo.logging import setup_logging

setup_logging()


class MimeoDTO:

    def __init__(self, source: dict):
        self.__source = source

    def __str__(self):
        return str(self.__source)


class MimeoConfig(MimeoDTO):

    OUTPUT_FORMAT_KEY = "output_format"
    OUTPUT_DETAILS_KEY = "output_details"
    OUTPUT_DETAILS_DIRECTION_KEY = "direction"
    OUTPUT_DETAILS_DIRECTORY_PATH_KEY = "directory_path"
    OUTPUT_DETAILS_FILE_NAME_KEY = "file_name"
    OUTPUT_DETAILS_METHOD_KEY = "method"
    OUTPUT_DETAILS_PROTOCOL_KEY = "protocol"
    OUTPUT_DETAILS_HOST_KEY = "host"
    OUTPUT_DETAILS_PORT_KEY = "port"
    OUTPUT_DETAILS_ENDPOINT_KEY = "endpoint"
    OUTPUT_DETAILS_AUTH_KEY = "auth"
    OUTPUT_DETAILS_USERNAME_KEY = "username"
    OUTPUT_DETAILS_PASSWORD_KEY = "password"
    XML_DECLARATION_KEY = "xml_declaration"
    INDENT_KEY = "indent"
    VARS_KEY = "vars"
    TEMPLATES_KEY = "_templates_"
    TEMPLATES_COUNT_KEY = "count"
    TEMPLATES_MODEL_KEY = "model"
    MODEL_CONTEXT_KEY = "context"
    MODEL_ATTRIBUTES_KEY = "_attrs"
    MODEL_VALUE_KEY = "_value"
    MODEL_MIMEO_UTIL_KEY = "_mimeo_util"
    MODEL_MIMEO_UTIL_NAME_KEY = "_name"

    OUTPUT_FORMAT_XML = "xml"

    OUTPUT_DETAILS_DIRECTION_FILE = "file"
    OUTPUT_DETAILS_DIRECTION_STD_OUT = "stdout"
    OUTPUT_DETAILS_DIRECTION_HTTP = "http"

    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST = "POST"
    OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_PUT = "PUT"

    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC = "basic"
    OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_DIGEST = "digest"

    SUPPORTED_OUTPUT_FORMATS = (OUTPUT_FORMAT_XML,)

    SUPPORTED_OUTPUT_DIRECTIONS = (OUTPUT_DETAILS_DIRECTION_STD_OUT,
                                   OUTPUT_DETAILS_DIRECTION_FILE,
                                   OUTPUT_DETAILS_DIRECTION_HTTP)
    SUPPORTED_REQUEST_METHODS = (OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST,
                                 OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_PUT)
    SUPPORTED_AUTH_METHODS = (OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC,
                              OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_DIGEST)
    REQUIRED_HTTP_DETAILS = (OUTPUT_DETAILS_HOST_KEY,
                             OUTPUT_DETAILS_ENDPOINT_KEY,
                             OUTPUT_DETAILS_USERNAME_KEY,
                             OUTPUT_DETAILS_PASSWORD_KEY)

    def __init__(self, config: dict):
        super().__init__(config)
        self.output_format = MimeoConfig.__get_output_format(config)
        self.output_details = MimeoOutputDetails(self.output_format, config.get(self.OUTPUT_DETAILS_KEY, {}))
        self.xml_declaration = config.get(self.XML_DECLARATION_KEY, False)
        self.indent = MimeoConfig.__get_indent(config)
        self.vars = MimeoConfig.__get_vars(config)
        self.templates = MimeoConfig.__get_templates(config)

    @staticmethod
    def __get_output_format(config):
        output_format = config.get(MimeoConfig.OUTPUT_FORMAT_KEY, MimeoConfig.OUTPUT_FORMAT_XML)
        if output_format in MimeoConfig.SUPPORTED_OUTPUT_FORMATS:
            return output_format
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_FORMAT_KEY,
                                           output_format,
                                           MimeoConfig.SUPPORTED_OUTPUT_FORMATS)

    @staticmethod
    def __get_indent(config):
        indent = config.get(MimeoConfig.INDENT_KEY, 0)
        if indent >= 0:
            return indent
        else:
            raise InvalidIndent(f"Provided indent [{indent}] is negative!")

    @staticmethod
    def __get_vars(config):
        variables = config.get(MimeoConfig.VARS_KEY, {})
        if not isinstance(variables, dict):
            raise InvalidVars(f"vars property does not store an object: {variables}")
        for var, val in variables.items():
            if not re.match(r"^[A-Z][A-Z_0-9]*$", var):
                raise InvalidVars(f"Provided var [{var}] is invalid "
                                  "(you can use upper-cased name with underscore and digits, starting with a letter)!")
            if isinstance(val, list) or (isinstance(val, dict) and not MimeoConfig.__is_mimeo_util_object(val)):
                raise InvalidVars(f"Provided var [{var}] is invalid (you can use ony atomic values and Mimeo Utils)!")
        return variables

    @staticmethod
    def __get_templates(config):
        templates = config.get(MimeoConfig.TEMPLATES_KEY)
        if templates is None:
            raise IncorrectMimeoConfig(f"No templates in the Mimeo Config: {config}")
        elif not isinstance(templates, list):
            raise IncorrectMimeoConfig(f"_templates_ property does not store an array: {config}")
        else:
            return (MimeoTemplate(template) for template in config.get(MimeoConfig.TEMPLATES_KEY))

    @staticmethod
    def __is_mimeo_util_object(obj: dict):
        return isinstance(obj, dict) and len(obj) == 1 and MimeoConfig.MODEL_MIMEO_UTIL_KEY in obj


class MimeoOutputDetails(MimeoDTO):

    def __init__(self, output_format: str, output_details: dict):
        super().__init__(output_details)
        self.direction = MimeoOutputDetails.__get_direction(output_details)
        MimeoOutputDetails.__validate_output_details(self.direction, output_details)
        self.directory_path = MimeoOutputDetails.__get_directory_path(self.direction, output_details)
        self.file_name_tmplt = MimeoOutputDetails.__get_file_name_tmplt(self.direction, output_details, output_format)
        self.method = MimeoOutputDetails.__get_method(self.direction, output_details)
        self.protocol = MimeoOutputDetails.__get_protocol(self.direction, output_details)
        self.host = MimeoOutputDetails.__get_host(self.direction, output_details)
        self.port = MimeoOutputDetails.__get_port(self.direction, output_details)
        self.endpoint = MimeoOutputDetails.__get_endpoint(self.direction, output_details)
        self.auth = MimeoOutputDetails.__get_auth(self.direction, output_details)
        self.username = MimeoOutputDetails.__get_username(self.direction, output_details)
        self.password = MimeoOutputDetails.__get_password(self.direction, output_details)

    @staticmethod
    def __get_direction(output_details):
        direction = output_details.get(MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY,
                                       MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE)
        if direction in MimeoConfig.SUPPORTED_OUTPUT_DIRECTIONS:
            return direction
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY,
                                           direction,
                                           MimeoConfig.SUPPORTED_OUTPUT_DIRECTIONS)

    @staticmethod
    def __validate_output_details(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            missing_details = []
            for detail in MimeoConfig.REQUIRED_HTTP_DETAILS:
                if detail not in output_details:
                    missing_details.append(detail)
            if len(missing_details) > 0:
                missing_details_str = ', '.join(missing_details)
                raise MissingRequiredProperty(f"Missing required fields is HTTP output details: {missing_details_str}")

    @staticmethod
    def __get_directory_path(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_DIRECTORY_PATH_KEY, "mimeo-output")

    @staticmethod
    def __get_file_name_tmplt(direction: str, output_details: dict, output_format: str):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE:
            file_name = output_details.get(MimeoConfig.OUTPUT_DETAILS_FILE_NAME_KEY, "mimeo-output")
            return f"{file_name}-{'{}'}.{output_format}"

    @staticmethod
    def __get_method(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            method = output_details.get(MimeoConfig.OUTPUT_DETAILS_METHOD_KEY,
                                        MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP_REQUEST_POST)
            if method in MimeoConfig.SUPPORTED_REQUEST_METHODS:
                return method
            else:
                raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_METHOD_KEY,
                                               method,
                                               MimeoConfig.SUPPORTED_REQUEST_METHODS)

    @staticmethod
    def __get_protocol(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PROTOCOL_KEY,
                                      MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP)

    @staticmethod
    def __get_host(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_HOST_KEY)

    @staticmethod
    def __get_port(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PORT_KEY)

    @staticmethod
    def __get_endpoint(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_ENDPOINT_KEY)

    @staticmethod
    def __get_auth(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            auth =  output_details.get(MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                                       MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP_AUTH_BASIC)
            if auth in MimeoConfig.SUPPORTED_AUTH_METHODS:
                return auth
            else:
                raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_AUTH_KEY,
                                               auth,
                                               MimeoConfig.SUPPORTED_AUTH_METHODS)

    @staticmethod
    def __get_username(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_USERNAME_KEY)

    @staticmethod
    def __get_password(direction: str, output_details: dict):
        if direction == MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP:
            return output_details.get(MimeoConfig.OUTPUT_DETAILS_PASSWORD_KEY)


class MimeoTemplate(MimeoDTO):

    def __init__(self, template: dict):
        super().__init__(template)
        MimeoTemplate.__validate_template(template)
        self.count = template.get(MimeoConfig.TEMPLATES_COUNT_KEY)
        self.model = MimeoModel(template.get(MimeoConfig.TEMPLATES_MODEL_KEY))

    @staticmethod
    def __validate_template(template: dict):
        if MimeoConfig.TEMPLATES_COUNT_KEY not in template:
            raise IncorrectMimeoTemplate(f"No count value in the Mimeo Template: {template}")
        if MimeoConfig.TEMPLATES_MODEL_KEY not in template:
            raise IncorrectMimeoTemplate(f"No model data in the Mimeo Template: {template}")


class MimeoModel(MimeoDTO):

    def __init__(self, model: dict):
        super().__init__(model)
        self.root_name = MimeoModel.__get_root_name(model)
        self.root_data = model.get(self.root_name)
        self.context_name = MimeoModel.__get_context_name(model, self.root_name)

    @staticmethod
    def __get_root_name(model: dict) -> str:
        model_keys = [key for key in filter(MimeoModel.__is_not_metadata_key, iter(model))]
        if len(model_keys) == 1:
            return model_keys[0]
        if len(model_keys) == 0:
            raise IncorrectMimeoModel(f"No root data in Mimeo Model: {model}")
        elif len(model_keys) > 1:
            raise IncorrectMimeoModel(f"Multiple root data in Mimeo Model: {model}")

    @staticmethod
    def __is_not_metadata_key(dict_key: str) -> bool:
        return dict_key not in [MimeoConfig.MODEL_CONTEXT_KEY]

    @staticmethod
    def __get_context_name(model: dict, root_name: str) -> str:
        context_name = model.get(MimeoConfig.MODEL_CONTEXT_KEY, root_name)
        if isinstance(context_name, str):
            return context_name
        else:
            raise IncorrectMimeoModel(f"Invalid context name in Mimeo Model (not a string value): {model}")



import logging
import re

from mimeo.config import MimeoConfig
from mimeo.context import MimeoContext, MimeoContextManager
from mimeo.context.annotations import mimeo_context
from mimeo.context.exc import VarNotFound
from mimeo.utils import (AutoIncrementUtil, CityUtil, CountryUtil,
                         CurrentIterationUtil, DateTimeUtil, DateUtil,
                         FirstNameUtil, KeyUtil, LastNameUtil, MimeoUtil,
                         RandomIntegerUtil, RandomItemUtil, RandomStringUtil)
from mimeo.utils.exc import InvalidMimeoUtil, InvalidValue, NotASpecialField

logger = logging.getLogger(__name__)


class UtilsRenderer:

    MIMEO_UTIL_NAME = "_name"
    MIMEO_UTILS = {
        RandomStringUtil.KEY: RandomStringUtil,
        RandomIntegerUtil.KEY: RandomIntegerUtil,
        RandomItemUtil.KEY: RandomItemUtil,
        DateUtil.KEY: DateUtil,
        DateTimeUtil.KEY: DateTimeUtil,
        AutoIncrementUtil.KEY: AutoIncrementUtil,
        CurrentIterationUtil.KEY: CurrentIterationUtil,
        KeyUtil.KEY: KeyUtil,
        CityUtil.KEY: CityUtil,
        CountryUtil.KEY: CountryUtil,
        FirstNameUtil.KEY: FirstNameUtil,
        LastNameUtil.KEY: LastNameUtil,
    }
    _INSTANCES = {}

    @classmethod
    def render_raw(cls, mimeo_util_key: str):
        return cls.render_parametrized({MimeoConfig.MODEL_MIMEO_UTIL_NAME_KEY: mimeo_util_key})

    @classmethod
    def render_parametrized(cls, mimeo_util_config: dict):
        logger.fine(f"Rendering a mimeo util [{mimeo_util_config}]")
        mimeo_util = cls._get_mimeo_util(mimeo_util_config)
        return mimeo_util.render()

    @classmethod
    def _get_mimeo_util(cls, mimeo_util_config: dict) -> MimeoUtil:
        mimeo_util_name = mimeo_util_config.get(MimeoConfig.MODEL_MIMEO_UTIL_NAME_KEY)
        if mimeo_util_name is None:
            raise InvalidMimeoUtil(f"Missing Mimeo Util name in configuration [{mimeo_util_config}]!")
        elif mimeo_util_name not in cls.MIMEO_UTILS:
            raise InvalidMimeoUtil(f"No such Mimeo Util [{mimeo_util_name}]!")

        mimeo_util_key = cls._generate_key(mimeo_util_config)
        if mimeo_util_key not in cls._INSTANCES:
            mimeo_util = cls._instantiate_mimeo_util(mimeo_util_name, mimeo_util_key, mimeo_util_config)
        else:
            mimeo_util = cls._INSTANCES.get(mimeo_util_key)

        return mimeo_util

    @staticmethod
    def _generate_key(mimeo_util_config: dict) -> str:
        return "-".join(":".join([key, str(val)]) for key, val in mimeo_util_config.items())

    @classmethod
    def _instantiate_mimeo_util(cls, name: str, key: str, config: dict) -> MimeoUtil:
        mimeo_util = cls.MIMEO_UTILS.get(name)(**config)
        cls._INSTANCES[key] = mimeo_util
        return mimeo_util


class VarsRenderer:

    @classmethod
    def render(cls, var: str):
        logger.fine(f"Rendering a variable [{var}]")
        return MimeoContextManager().get_var(var)


class SpecialFieldsRenderer:

    @classmethod
    @mimeo_context
    def render(cls, field_name: str, context: MimeoContext = None):
        logger.fine(f"Rendering a special field [{field_name}]")
        return context.curr_iteration().get_special_field(field_name)


class MimeoRenderer:

    _UTILS_PATTERN = re.compile("^{(.+)}$")
    _VARS_PATTERN = re.compile(".*({[A-Z_0-9]+})")
    _SPECIAL_FIELDS_PATTERN = re.compile(".*({:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:})")

    @classmethod
    def get_special_field_name(cls, field_name: str) -> str:
        if not cls.is_special_field(field_name):
            raise NotASpecialField(f"Provided field [{field_name}] is not a special one (use {'{:NAME:}'})!")

        return field_name[2:][:-2]

    @classmethod
    def is_special_field(cls, field_name: str) -> bool:
        return bool(re.match(r"^{:([a-zA-Z]+:)?[-_a-zA-Z0-9]+:}$", field_name))

    @classmethod
    def is_raw_mimeo_util(cls, value: str):
        raw_mimeo_utils = UtilsRenderer.MIMEO_UTILS.keys()
        raw_mimeo_utils_re = "^{(" + "|".join(raw_mimeo_utils) + ")}$"
        return bool(re.match(raw_mimeo_utils_re, value))

    @classmethod
    def is_parametrized_mimeo_util(cls, value: dict):
        return isinstance(value, dict) and len(value) == 1 and MimeoConfig.MODEL_MIMEO_UTIL_KEY in value

    @classmethod
    def render(cls, value):
        logger.fine(f"Rendering a value [{value}]")
        try:
            if isinstance(value, str):
                return cls._render_string_value(value)
            elif cls.is_parametrized_mimeo_util(value):
                return cls._render_parametrized_mimeo_util(value)
            else:
                return value
        except (InvalidMimeoUtil, VarNotFound, InvalidValue) as err:
            error_name = type(err).__name__
            logger.error(f"The [{error_name}] error occurred during rendering a value [{value}]: [{err}].")
            raise err

    @classmethod
    def _render_string_value(cls, value: str):
        if cls._SPECIAL_FIELDS_PATTERN.match(value):
            return cls._render_special_field(value)
        elif cls._VARS_PATTERN.match(value):
            return cls._render_var(value)
        elif cls.is_raw_mimeo_util(value):
            return cls._render_raw_mimeo_util(value)
        else:
            return value

    @classmethod
    def _render_special_field(cls, value):
        match = next(cls._SPECIAL_FIELDS_PATTERN.finditer(value))
        wrapped_special_field = match.group(1)
        rendered_value = SpecialFieldsRenderer.render(wrapped_special_field[2:][:-2])
        logger.fine(f"Rendered special field value [{rendered_value}]")
        if len(wrapped_special_field) != len(value):
            rendered_value = str(rendered_value).lower() if isinstance(rendered_value, bool) else str(rendered_value)
            rendered_value = value.replace(wrapped_special_field, str(rendered_value))
        return cls.render(rendered_value)

    @classmethod
    def _render_var(cls, value):
        match = next(cls._VARS_PATTERN.finditer(value))
        wrapped_var = match.group(1)
        rendered_value = VarsRenderer.render(wrapped_var[1:][:-1])
        logger.fine(f"Rendered variable value [{rendered_value}]")
        if cls.is_parametrized_mimeo_util(rendered_value):
            rendered_value = cls._render_parametrized_mimeo_util(rendered_value)
        if len(wrapped_var) != len(value):
            rendered_value = str(rendered_value).lower() if isinstance(rendered_value, bool) else str(rendered_value)
            rendered_value = value.replace(wrapped_var, str(rendered_value))
        return cls.render(rendered_value)

    @classmethod
    def _render_raw_mimeo_util(cls, value):
        rendered_value = UtilsRenderer.render_raw(value[1:][:-1])
        return cls.render(rendered_value)

    @classmethod
    def _render_parametrized_mimeo_util(cls, value: dict):
        mimeo_util = value[MimeoConfig.MODEL_MIMEO_UTIL_KEY]
        mimeo_util = cls._render_mimeo_util_parameters(mimeo_util)
        logger.fine(f"Pre-rendered mimeo util [{mimeo_util}]")
        rendered_value = UtilsRenderer.render_parametrized(mimeo_util)
        return cls.render(rendered_value)

    @classmethod
    def _render_mimeo_util_parameters(cls, mimeo_util_config: dict) -> dict:
        logger.fine("Rendering mimeo util parameters")
        return {key: cls.render(value) if key != "_name" else value for key, value in mimeo_util_config.items()}



import logging

from requests import Response, Session
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from mimeo.config.mimeo_config import MimeoOutputDetails
from mimeo.consumers import Consumer

logger = logging.getLogger(__name__)


class HttpConsumer(Consumer):

    def __init__(self, output_details: MimeoOutputDetails):
        self.method = output_details.method
        self.url = HttpConsumer.__build_url(output_details)
        if output_details.auth == "basic":
            self.__auth = HTTPBasicAuth(output_details.username, output_details.password)
        else:
            self.__auth = HTTPDigestAuth(output_details.username, output_details.password)

    def consume(self, data: str) -> Response:
        logger.fine(f"Consuming data [{data}]")
        with Session() as sess:
            logger.info(f"Sending request {self.method} {self.url}")
            return sess.request(self.method,
                                self.url,
                                auth=self.__auth,
                                data=data,
                                headers={"Content-Type": "application/xml"})

    @staticmethod
    def __build_url(output_details: MimeoOutputDetails):
        if output_details.port is None:
            return f"{output_details.protocol}://{output_details.host}{output_details.endpoint}"
        else:
            return f"{output_details.protocol}://{output_details.host}:{output_details.port}{output_details.endpoint}"



from mimeo.config.exc import UnsupportedPropertyValue
from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import Consumer, FileConsumer, HttpConsumer, RawConsumer


class ConsumerFactory:

    FILE_DIRECTION = MimeoConfig.OUTPUT_DETAILS_DIRECTION_FILE
    STD_OUT_DIRECTION = MimeoConfig.OUTPUT_DETAILS_DIRECTION_STD_OUT
    HTTP_DIRECTION = MimeoConfig.OUTPUT_DETAILS_DIRECTION_HTTP

    @staticmethod
    def get_consumer(mimeo_config: MimeoConfig) -> Consumer:
        direction = mimeo_config.output_details.direction
        if direction == ConsumerFactory.STD_OUT_DIRECTION:
            return RawConsumer()
        elif direction == ConsumerFactory.FILE_DIRECTION:
            return FileConsumer(mimeo_config.output_details)
        elif direction == ConsumerFactory.HTTP_DIRECTION:
            return HttpConsumer(mimeo_config.output_details)
        else:
            raise UnsupportedPropertyValue(MimeoConfig.OUTPUT_DETAILS_DIRECTION_KEY,
                                           direction,
                                           MimeoConfig.SUPPORTED_OUTPUT_DIRECTIONS)



import logging
import xml.etree.ElementTree as ElemTree
from typing import Iterator, List, Union
from xml.dom import minidom

from mimeo.config.mimeo_config import MimeoConfig, MimeoTemplate
from mimeo.context import MimeoContext
from mimeo.context.annotations import (mimeo_clear_iterations, mimeo_context,
                                       mimeo_context_switch,
                                       mimeo_next_iteration)
from mimeo.generators import Generator
from mimeo.utils import MimeoRenderer

logger = logging.getLogger(__name__)


class XMLGenerator(Generator):

    def __init__(self, mimeo_config: MimeoConfig):
        super().__init__()
        self.__indent = mimeo_config.indent
        self.__xml_declaration = mimeo_config.xml_declaration

    @classmethod
    def generate(cls, templates: Union[list, Iterator[MimeoTemplate]], parent: ElemTree.Element = None) -> Iterator[ElemTree.Element]:
        for template in templates:
            for copy in cls._process_single_template(template, parent):
                yield copy

    def stringify(self, root, mimeo_config):
        if self.__indent is None or self.__indent == 0:
            return ElemTree.tostring(root,
                                     encoding="utf-8",
                                     method="xml",
                                     xml_declaration=self.__xml_declaration).decode('ascii')
        else:
            xml_string = ElemTree.tostring(root)
            xml_minidom = minidom.parseString(xml_string)
            if self.__xml_declaration:
                return xml_minidom.toprettyxml(indent=" " * self.__indent, encoding="utf-8").decode('ascii')
            else:
                return xml_minidom.childNodes[0].toprettyxml(indent=" " * self.__indent, encoding="utf-8").decode('ascii')

    @classmethod
    @mimeo_context_switch
    @mimeo_clear_iterations
    def _process_single_template(cls, template: MimeoTemplate, parent: ElemTree.Element = None) -> List[ElemTree.Element]:
        logger.debug(f"Reading template [{template}]")
        copies = [cls._process_single_copy(template, parent) for _ in iter(range(template.count))]
        return copies

    @classmethod
    @mimeo_next_iteration
    def _process_single_copy(cls, template: MimeoTemplate, parent: ElemTree.Element = None):
        return cls._process_node(parent, template.model.root_name, template.model.root_data)

    @classmethod
    @mimeo_context
    def _process_node(cls, parent, element_tag, element_value, attributes: dict = None, context: MimeoContext = None):
        logger.fine(f"Rendering element - "
                    f"parent [{parent if parent is None else parent.tag}], "
                    f"element_tag [{element_tag}], "
                    f"element_value [{element_value}], "
                    f"attributes [{attributes}]")
        attributes = attributes if attributes is not None else {}
        if element_tag == MimeoConfig.TEMPLATES_KEY:
            templates = (MimeoTemplate(template) for template in element_value)
            for _ in cls.generate(templates, parent):
                pass
        else:
            is_special_field = MimeoRenderer.is_special_field(element_tag)
            if is_special_field:
                element_tag = MimeoRenderer.get_special_field_name(element_tag)

            element = cls._create_xml_element(parent, element_tag, attributes)

            if isinstance(element_value, dict) and not MimeoRenderer.is_parametrized_mimeo_util(element_value):
                if MimeoConfig.MODEL_ATTRIBUTES_KEY in element_value:
                    element_value_copy = dict(element_value)
                    attrs = element_value_copy.pop(MimeoConfig.MODEL_ATTRIBUTES_KEY)
                    value = element_value_copy.get(MimeoConfig.MODEL_VALUE_KEY, element_value_copy)
                    if parent is not None:
                        parent.remove(element)
                        cls._process_node(parent, element_tag, value, attrs)
                    else:
                        return cls._process_node(parent, element_tag, value, attrs)
                else:
                    for child_tag, child_value in element_value.items():
                        cls._process_node(element, child_tag, child_value)
            elif isinstance(element_value, list):
                has_only_atomic_values = all(not isinstance(child, (list, dict)) for child in element_value)
                if has_only_atomic_values:
                    parent.remove(element)
                    for child in element_value:
                        cls._process_node(parent, element_tag, child)
                else:
                    for child in element_value:
                        grand_child_tag = next(iter(child))
                        grand_child_data = child[grand_child_tag]
                        cls._process_node(element, grand_child_tag, grand_child_data)
            else:
                value = MimeoRenderer.render(element_value)
                if is_special_field:
                    context.curr_iteration().add_special_field(element_tag, value)

                value_str = str(value) if value is not None else ""
                element.text = value_str.lower() if isinstance(value, bool) else value_str
                logger.fine(f"Rendered value [{element.text}]")

            if parent is None:
                return element

    @classmethod
    def _create_xml_element(cls, parent, element_tag, attributes):
        if parent is None:
            return ElemTree.Element(element_tag, attrib=attributes)
        else:
            return ElemTree.SubElement(parent, element_tag, attrib=attributes)



class UnsupportedPropertyValue(Exception):

    def __init__(self, prop: str, val: str, supported_values: tuple):
        super().__init__(f"Provided {prop} [{val}] is not supported! "
                         f"Supported values: [{', '.join(supported_values)}].")


class MissingRequiredProperty(Exception):
    pass


class InvalidIndent(Exception):
    pass


class InvalidVars(Exception):
    pass


class IncorrectMimeoModel(Exception):
    pass


class IncorrectMimeoTemplate(Exception):
    pass


class IncorrectMimeoConfig(Exception):
    pass
